from typing import List, Tuple
from wsgiref.simple_server import make_server

from . import methods
from . import response
from .requestcontext import RequestContext, HandlerFunc
from .router import Router


class RouterGroup:
    engine: "Engine"
    parent: str
    prefix: str
    middlewares: Tuple["HandlerFunc"]

    def __init__(
        self,
        engine: "Engine" = None,
        prefix="",
        parent=None,
        *middlewares: "HandlerFunc",
    ):
        self.prefix = prefix
        self.middlewares = middlewares
        self.parent = parent
        self.engine = engine

    def group(self, prefix, *middlewares: "HandlerFunc"):
        """
        create router group
        """
        new_group = RouterGroup(
            prefix=self.prefix + prefix, parent=self, engine=self.engine
        )
        new_group.middlewares += middlewares
        self.engine.groups.append(new_group)
        return new_group

    def add_route(self, method, pattern, *handlers: HandlerFunc):
        """
        add a pattern to router
        """
        self.engine.router.add_route(method, self.prefix + pattern, *handlers)

    def get(self, pattern, *handlers):
        self.add_route(methods.GET, pattern, *handlers)

    def head(self, pattern, *handlers):
        self.add_route(methods.HEAD, pattern, *handlers)

    def post(self, pattern, *handlers):
        self.add_route(methods.POST, pattern, *handlers)

    def put(self, pattern, *handlers):
        self.add_route(methods.PUT, pattern, *handlers)

    def delete(self, pattern, *handlers):
        self.add_route(methods.DELETE, pattern, *handlers)

    def connect(self, pattern, *handlers):
        self.add_route(methods.CONNECT, pattern, *handlers)

    def options(self, pattern, *handlers):
        self.add_route(methods.OPTIONS, pattern, *handlers)

    def trace(self, pattern, *handlers):
        self.add_route(methods.TRACE, pattern, *handlers)

    def patch(self, pattern, *handlers):
        self.add_route(methods.PATCH, pattern, *handlers)

    def all(self, pattern, *handlers):
        for method in methods.ALL:
            self.add_route(method, pattern, *handlers)

    def no_route(self, *handlers):
        """
        set no route handler
        """
        self.engine.no_route_handler = list(handlers)

    def no_method(self, *handlers):
        """
        set no method handler
        """
        self.engine.no_method_handler = list(handlers)

    def use(self, *middlewares):
        """
        use middlewares before handlers
        """
        self.middlewares += middlewares


class Engine(RouterGroup):
    router_group: "RouterGroup"
    router: "Router"
    prefix: str
    middlewares = []
    groups: List["RouterGroup"]
    engine: "Engine"
    no_route_handler: List["HandlerFunc"]
    no_method_handler: List["HandlerFunc"]

    def __init__(self, re=False, base=""):
        self.router_group = RouterGroup(engine=self)
        self.engine = self
        self.no_method_handler = [response.error_method_not_allow]
        self.no_route_handler = [response.error_not_found]
        self.prefix = "" if base == "/" else base
        self.groups = []
        self.router = Router()
        if re:
            self.use_regex()

    def use_regex(self):
        self.router.re = True

    def run(
        self,
        port: int,
        addr="localhost",
        fmt="Running on http://{addr}:{port}",
        poll_interval: float = 0.5,
    ):
        """
        start a http server
        """
        httpd = make_server(addr, port, self.serve_http)
        print(fmt.format(addr=addr, port=port))
        httpd.serve_forever(poll_interval=poll_interval)

    def serve_http(self, env, start_response):
        """
        serve http request
        """
        middlewares = self.middlewares
        ctx = RequestContext(env, self.engine)
        path = ctx.path()
        prefix = ""

        for group in self.groups:
            if path.startswith(group.prefix + "/"):
                prefix = group.prefix
                middlewares += group.middlewares

        ctx.handlers = list(middlewares)
        if self.router.re:
            handle_response = self.router.handle_prefix(ctx, prefix)
        else:
            handle_response = self.router.handle(ctx)

        return handle_response.start_response(start_response)



import re
from types import FunctionType
from typing import Dict, List, Optional

from .node import Node, wild_of
from .requestcontext import RequestContext
from .response import Response


def parse_pattern(pattern: str) -> List[str]:
    vs = pattern.split("/")
    parts: List[str] = []
    for item in vs:
        if item != "":
            parts.append(item)
            if item[0] == "*":
                break
    return parts


def split_pattern(s: str) -> List[str]:
    return parse_pattern(s)


def split_slash(s: str) -> List[str]:
    vs = s.split("/")
    parts: List[str] = []
    for item in vs:
        if item != "":
            parts.append(item)
    return parts


class Router:
    root: Node
    handlers: Dict[str, Dict[str, List[FunctionType]]]
    re: bool

    def __init__(self):
        self.root = Node()
        self.handlers = {}
        self.re = False

    def add_route(self, method: str, pattern: str, *handler_func: FunctionType):
        """
        add pattern to router

        method will be upper

        expand handler functions to handlers[pattern][method.upper()]
        """
        if not self.re:
            self.root.set(pattern)
        if pattern not in self.handlers:
            self.handlers[pattern] = {}
        self.handlers[pattern][method.upper()] = list(handler_func)

    def get_route(self, path: str) -> tuple[Optional["Node"], Optional[dict]]:
        """
        parse path to pattern (split "/")

        get node and dynamic field mapping
        """
        search_parts = split_pattern(path)
        node = self.root.get(path)
        if node:
            params = {}
            parts = split_pattern(node.pattern)
            for index, part in enumerate(parts):
                is_wild, wild_key, is_multi = wild_of(part)
                if is_wild:
                    if is_multi:
                        params[wild_key] = "/".join(search_parts[index:])
                        break
                    else:
                        params[wild_key] = search_parts[index]

            return node, params

        return None, None

    def handle(self, ctx: RequestContext) -> Response:
        """
        handle a request

        if no route defined, handle engine.no_route_handler

        if route defined, no method defined, handle engine.no_method_handler

        :param ctx: request context
        :return: response handler's response
        """

        node, params = self.get_route(ctx.path())
        ctx.params = params
        if node is not None:
            # has router
            pattern = node.pattern
            method = ctx.method()

            h_p = self.handlers.get(pattern)
            # if h_p:
            h_p_m = h_p.get(method)
            if h_p_m:
                # has method
                ctx.handlers.extend(h_p_m)
            else:
                # no method
                ctx.handlers.extend(ctx.engine.no_method_handler)
        else:
            # no route
            ctx.handlers.extend(ctx.engine.no_route_handler)

        return ctx.next()

    def handle_prefix(self, ctx: RequestContext, prefix: str) -> Response:
        """
        handle a request

        if no route defined, handle engine.no_route_handler

        if route defined, no method defined, handle engine.no_method_handler

        :param ctx: request
        :param prefix: group prefix
        :return: response handler's response
        """
        path = ctx.path()
        method = ctx.method()
        for regex in self.handlers:
            if len(regex) < len(prefix) or len(regex[: len(prefix)]) < len(prefix):
                continue
            if re.match(regex[len(regex):], path[len(regex):]):
                key = prefix + regex
                handlers_map = self.handlers.get(key)
                handlers = handlers_map.get(method)
                if handlers is None:
                    ctx.handlers.extend(ctx.engine.no_method_handler)
                else:
                    ctx.handlers.extend(handlers)
                return ctx.next()

        ctx.handlers.extend(ctx.engine.no_route_handler)
        return ctx.next()



import urllib.parse
from typing import List, Optional, TYPE_CHECKING

from .response import Response
from .utils import execute_func

if TYPE_CHECKING:
    from .engine import Engine


class HandlerFunc:
    def __call__(
            self, _1: Optional["Request"], _2: Optional["Response"]
    ) -> Optional[Response]:
        pass


class RequestContext:
    handlers: List["HandlerFunc"]
    _index: int
    response: Response
    engine: "Engine"
    params: Optional[dict]

    def __init__(self, environ, engine=None):
        self.engine = engine
        self._environ = environ
        self._index = -1
        self.handlers = []

    def next(self) -> Response:
        """
        increase index, call next handler function
        :return: response
        """
        self._index += 1
        self.response = Response()
        while self._index < len(self.handlers):
            next_func = self.handlers[self._index]
            has_return = execute_func(next_func, self, self.response)
            if has_return:
                # has return, response is from return value
                if isinstance(has_return, Response):
                    # response is Response
                    self.response = has_return
                else:
                    # encode response to Response
                    self.response = Response(str(has_return).encode())
            self._index += 1
        return self.response

    def abort(self):
        """
        abort at current handler
        """
        self._index = len(self.handlers)

    def reset(self):
        """
        reset handlers index
        """
        self._index = -1

    def method(self) -> str:
        """
        get http method
        """
        return self._environ["REQUEST_METHOD"]

    def server_port(self) -> str:
        """
        get port
        """
        return self._environ["SERVER_PORT"]

    def path(self) -> str:
        """
        get url path
        """
        return self._environ["PATH_INFO"]

    def query_string(self):
        """
        get query string
        """
        return self._environ["QUERY_STRING"]

    def content_length(self):
        """
        get content-length
        """
        return self._environ["CONTENT_LENGTH"]

    def content_type(self):
        """
        get content-type
        """
        return self._environ["CONTENT_TYPE"]

    def headers(self):
        """
        get headers
        """
        return self._environ["HTTP_ACCEPT"]

    @property
    def query(self):
        """
        get query arguments
        """
        get_arguments = urllib.parse.parse_qs(self._environ["QUERY_STRING"])
        return {k: v[0] for k, v in get_arguments.items()}


class Request(RequestContext):
    pass


"""
alias of RequestContext
"""



import json
from wsgiref.headers import Headers
from .utils import http_status_text


class Response:
    response: bytes
    _status: int
    charset: str
    headers: "Headers"

    def __init__(
        self, response: bytes = None, status=200, charset="utf-8", header=None
    ):
        self.response = response
        self.charset = charset
        self.headers = Headers() if header is None else header
        if charset:
            self.headers.add_header("charset", charset)
        self._status = status

    def status(self, code):
        """
        set http status

        :param code: status
        """
        self._status = code

    def write(self, data: bytes):
        """
        set response data (bytes)

        :param data: response data
        """
        self.response = data

    def string(self, s: str):
        """
        set response data (format string)

        :param s: format string
        """
        self.write(s.encode(self.charset))

    def json(self, data, **params):
        """
        set response data (json)
        """
        self.content_type("application/json")
        self.string(json.dumps(data, **params))

    def set_header(self, key: str, value: str):
        self.headers.add_header(key, value)

    def content_type(self, content_type: str):
        """
        add header Content-Type
        """
        self.set_header("content-type", content_type)

    def content_length(self, length: int):
        """
        add header Content-Length
        """
        self.set_header("content-length", length)

    def start_response(self, start_response):
        start_response(
            f"{self._status} {http_status_text(self._status)}", self.headers.items()
        )
        if self.response:
            if isinstance(self.response, bytes):
                yield self.response
            else:
                yield self.response.encode(self.charset)


def error_not_found() -> Response:
    return response_status_text(404)


def error_method_not_allow() -> Response:
    return response_status_text(405)


def response_status_text(code=200) -> Response:
    return Response(http_status_text(code).encode(), code)



"""
http methods
"""

GET = "GET"
HEAD = "HEAD"
POST = "POST"
PUT = "PUT"
DELETE = "DELETE"
CONNECT = "CONNECT"
OPTIONS = "OPTIONS"
TRACE = "TRACE"
PATCH = "PATCH"

ALL = [GET, HEAD, POST, PUT, DELETE, CONNECT, OPTIONS, TRACE, PATCH]



import http
from types import FunctionType


def execute_func(func: FunctionType, *args, **kwargs) -> FunctionType:
    args_needed = func.__code__.co_argcount

    if args_needed > len(args):
        args += (None,) * (args_needed - len(args))

    if args_needed < len(args):
        args = args[:args_needed]

    return func(*args, **kwargs)


def http_status_text(status) -> str:
    """
    return status' phrase
    """
    return http.HTTPStatus(status).phrase



from types import FunctionType
from typing import List
from ..engine import RouterGroup
from .. import methods
from ..requestcontext import HandlerFunc


def handler(pattern: str, methods_list: List[str], group: "RouterGroup"):
    """
    wrap a handler call add_route

    example
    ```py
    app = Engine()
    # function decorators
    @handler("/", ["GET"], app)
    def index_handler():
        pass
    # as
    app.add_route("GET", "/", index_handler)

    # class decorators
    @handler("/", ["GET"], app)
    class Pages:
        def middle_handler():
            pass
        def index_handler():
            pass
    # as
    engine.add_route("GET", Pages.middle_handler, Pages.index_handler)
    """

    def decorators(var: HandlerFunc | type):
        if group:
            for method in methods_list:
                handlers = ()

                if isinstance(var, FunctionType):
                    handlers = (var,)

                elif isinstance(var, type):
                    members = vars(var)
                    for member in members:
                        attr = getattr(var, member)
                        if callable(attr):
                            handlers += (attr,)

                if len(handlers):
                    group.add_route(method, pattern, *handlers)

    return decorators


def get(pattern: str, group: "RouterGroup"):
    """
    wrap a handler call get

    example
    ```
    @get("/")
    def index_handler(req, res):
        pass
    # as
    app.get("/", index_handler)
    ```
    """
    return handler(pattern, [methods.GET], group)


def head(pattern: str, group: "RouterGroup"):
    return handler(pattern, [methods.HEAD], group)


def post(pattern: str, group: "RouterGroup"):
    return handler(pattern, [methods.POST], group)


def put(pattern: str, group: "RouterGroup"):
    return handler(pattern, [methods.PUT], group)


def delete(pattern: str, group: "RouterGroup"):
    return handler(pattern, [methods.DELETE], group)


def connect(pattern: str, group: "RouterGroup"):
    return handler(pattern, [methods.CONNECT], group)


def options(pattern: str, group: "RouterGroup"):
    return handler(pattern, [methods.OPTIONS], group)


def trace(pattern: str, group: "RouterGroup"):
    return handler(pattern, [methods.TRACE], group)


def patch(pattern: str, group: "RouterGroup"):
    return handler(pattern, [methods.PATCH], group)


def all(pattern: str, group: "RouterGroup"):
    return handler(pattern, methods.ALL, group)



from functools import cmp_to_key
from typing import List, Optional


class Node:
    pattern: str
    part: str
    children: List["Node"]
    is_wild: bool

    def __init__(
            self,
            pattern: str = "",
            part: str = "",
            is_wild=False,
    ):
        self.pattern = pattern
        self.part = part
        self.children = []
        self.is_wild = is_wild

    def set(self, s: str):
        self.insert(s, split_pattern(s), 0)

    def get(self, s: str):
        return self.search(split_slash(s), 0)

    def insert(self, pattern: str, parts: List[str], height: int):
        if len(parts) == height:
            self.pattern = pattern
            return

        part = parts[height]
        is_wild, _, _ = wild_of(part)
        child = self.find_wild_child()
        if child and child.is_wild and is_wild:
            return

        child = self.find_specific_child(part)
        if child is None:
            child = Node(part=part, is_wild=is_wild)
            self.children.append(child)
        child.insert(pattern, parts, height + 1)
        self.sort()

    def search(self, parts: List[str], height: int) -> Optional["Node"]:
        _, _, is_multi = wild_of(self.part)
        if len(parts) == height or is_multi:
            if not self.pattern:
                return None
            return self

        part = parts[height]
        children = self.match_children(part)

        for child in children:
            result = child.search(parts, height + 1)
            if result:
                return result

        return None

    # def match_child(self, part: str) -> Optional["Node"]:
    #     for child in self.children:
    #         if child.part == part or child.is_wild:
    #             return child
    #     return None

    def find_wild_child(self):
        for child in self.children:
            if child.is_wild:
                return child
        return None

    def find_specific_child(self, part):
        for child in self.children:
            if child.part == part:
                return child
        return None

    def match_children(self, part: str) -> List["Node"]:
        nodes = []
        for child in self.children:
            if child.part == part or child.is_wild:
                nodes.append(child)
        return nodes

    def sort(self):
        self.children = sorted(self.children, key=cmp_to_key(cmp))
        for i in self.children:
            i.sort()


def wild_of(s):
    if not s:
        return False, "", False
    if s[0] == '*':
        return True, s[1:], True
    if s[0] == ':':
        return True, s[1:], False
    if s[0] == '{' and s[-1] == '}':
        return True, s[1:-1], False
    if s.startswith('...'):
        return True, s[3:], True
    return False, "", False


def split_pattern(pattern):
    parts = []
    for part in pattern.split("/"):
        if part:
            parts.append(part)
            _, _, is_multi = wild_of(part)
            if is_multi:
                break
    return parts


def split_slash(s):
    parts = []
    for part in s.split("/"):
        if part:
            parts.append(part)
    return parts


def cmp(i: Node, j: Node):
    if not i.is_wild and j.is_wild:
        return -1
    elif i.is_wild and not j.is_wild:
        return 1
    else:
        return len(i.pattern) - len(j.pattern)
